commit 659f197e185606ae5d4dea904e18f4e392a52f68
Author: Maxim Kuvyrkov <maxim@codesourcery.com>
Date:   Sat Oct 1 01:09:50 2011 -0700

    Add iterative optimization passes

diff --git a/gcc/Makefile.in b/gcc/Makefile.in
index 3608904..ac1839d 100644
--- a/gcc/Makefile.in
+++ b/gcc/Makefile.in
@@ -2657,7 +2657,7 @@ tree-optimize.o : tree-optimize.c $(TREE_FLOW_H) $(CONFIG_H) $(SYSTEM_H) \
    coretypes.h $(TREE_DUMP_H) toplev.h $(DIAGNOSTIC_CORE_H) $(FUNCTION_H) langhooks.h \
    $(FLAGS_H) $(CGRAPH_H) $(PLUGIN_H) \
    $(TREE_INLINE_H) tree-mudflap.h $(GGC_H) graph.h $(CGRAPH_H) \
-   $(TREE_PASS_H) $(CFGLOOP_H) $(EXCEPT_H) $(REGSET_H)
+   $(TREE_PASS_H) $(CFGLOOP_H) $(EXCEPT_H) $(REGSET_H) $(PARAMS_H)
 
 gimplify.o : gimplify.c $(CONFIG_H) $(SYSTEM_H) $(TREE_H) $(GIMPLE_H) \
    $(DIAGNOSTIC_H) $(GIMPLE_H) $(TREE_INLINE_H) langhooks.h \
diff --git a/gcc/cgraph.c b/gcc/cgraph.c
index f056d3d..4738b28 100644
--- a/gcc/cgraph.c
+++ b/gcc/cgraph.c
@@ -2416,7 +2416,7 @@ cgraph_add_new_function (tree fndecl, bool lowered)
 	    tree_lowering_passes (fndecl);
 	    bitmap_obstack_initialize (NULL);
 	    if (!gimple_in_ssa_p (DECL_STRUCT_FUNCTION (fndecl)))
-	      execute_pass_list (pass_early_local_passes.pass.sub);
+	      execute_early_local_passes_for_current_function ();
 	    bitmap_obstack_release (NULL);
 	    pop_cfun ();
 	    current_function_decl = NULL;
@@ -2441,7 +2441,7 @@ cgraph_add_new_function (tree fndecl, bool lowered)
 	gimple_register_cfg_hooks ();
 	bitmap_obstack_initialize (NULL);
 	if (!gimple_in_ssa_p (DECL_STRUCT_FUNCTION (fndecl)))
-	  execute_pass_list (pass_early_local_passes.pass.sub);
+	  execute_early_local_passes_for_current_function ();
 	bitmap_obstack_release (NULL);
 	tree_rest_of_compilation (fndecl);
 	pop_cfun ();
diff --git a/gcc/cgraphunit.c b/gcc/cgraphunit.c
index 25d7561..9a0e06d 100644
--- a/gcc/cgraphunit.c
+++ b/gcc/cgraphunit.c
@@ -255,7 +255,7 @@ cgraph_process_new_functions (void)
 	      /* When not optimizing, be sure we run early local passes anyway
 		 to expand OMP.  */
 	      || !optimize)
-	    execute_pass_list (pass_early_local_passes.pass.sub);
+	    execute_early_local_passes_for_current_function ();
 	  else
 	    compute_inline_parameters (node, true);
 	  free_dominance_info (CDI_POST_DOMINATORS);
diff --git a/gcc/doc/invoke.texi b/gcc/doc/invoke.texi
index 50e875a..d5cc035 100644
--- a/gcc/doc/invoke.texi
+++ b/gcc/doc/invoke.texi
@@ -9088,6 +9088,12 @@ the parameter is reserved exclusively for debug insns created by
 @option{-fvar-tracking-assignments}, but debug insns may get
 (non-overlapping) uids above it if the reserved range is exhausted.
 
+@item eipa-iterations
+The pass scheduler will execute @option{eipa-iterations} iterations of
+early optimization passes before running interprocedural analysis.
+Running several iterations of optimization passes allows the compiler
+to provide thoroughly optimized code to the interprocedural analysis.
+
 @item ipa-sra-ptr-growth-factor
 IPA-SRA will replace a pointer to an aggregate with one or more new
 parameters only when their cumulative size is less or equal to
diff --git a/gcc/params.def b/gcc/params.def
index b160530..2baad91 100644
--- a/gcc/params.def
+++ b/gcc/params.def
@@ -861,6 +861,11 @@ DEFPARAM (PARAM_MIN_NONDEBUG_INSN_UID,
 	  "The minimum UID to be used for a nondebug insn",
 	  0, 1, 0)
 
+DEFPARAM (PARAM_EIPA_ITERATIONS,
+	  "eipa-iterations",
+	  "Number of iterations of early optimizations before IPA analysis",
+	  3, 1, 0)
+
 DEFPARAM (PARAM_IPA_SRA_PTR_GROWTH_FACTOR,
 	  "ipa-sra-ptr-growth-factor",
 	  "Maximum allowed growth of size of new parameters ipa-sra replaces "
diff --git a/gcc/passes.c b/gcc/passes.c
index 887007f..c3caf3f 100644
--- a/gcc/passes.c
+++ b/gcc/passes.c
@@ -898,10 +898,13 @@ is_pass_explicitly_enabled_or_disabled (struct opt_pass *pass,
 }
 
 /* Look at the static_pass_number and duplicate the pass
-   if it is already added to a list. */
+   if it is already added to a list.
+   If SUFFIX is non-NULL, append it to the name of the new pass instead
+   of an increasing number.  */
 
 static struct opt_pass *
-make_pass_instance (struct opt_pass *pass, bool track_duplicates)
+make_pass_instance (struct opt_pass *pass, bool track_duplicates,
+		    const char *suffix)
 {
   /* A nonzero static_pass_number indicates that the
      pass is already in the list.  */
@@ -924,15 +927,40 @@ make_pass_instance (struct opt_pass *pass, bool track_duplicates)
       else
         gcc_unreachable ();
 
+      if (pass->sub)
+	/* Duplicate sub-passes.  */
+	{
+	  struct opt_pass *sub = pass->sub;
+	  struct opt_pass **new_sub_ptr = &new_pass->sub;
+
+	  do
+	    {
+	      *new_sub_ptr = make_pass_instance (sub, track_duplicates, suffix);
+	      new_sub_ptr = &(*new_sub_ptr)->next;
+	      sub = sub->next;
+	    } while (sub);
+	}
+
       new_pass->next = NULL;
 
       new_pass->todo_flags_start &= ~TODO_mark_first_instance;
 
+      if (suffix)
+	/* Make pass name unique by appending SUFFIX to it.  */
+	{
+	  /* Make sure we don't append suffix twice.  */
+	  {
+	    int name_len = strlen (pass->name), suffix_len = strlen (suffix);
+	    gcc_assert (strcmp (suffix, &pass->name[name_len - suffix_len]));
+	  }
+	  new_pass->static_pass_number = -1;
+	  new_pass->name = concat (pass->name, suffix, NULL);
+	}
       /* Indicate to register_dump_files that this pass has duplicates,
          and so it should rename the dump file.  The first instance will
          be -1, and be number of duplicates = -static_pass_number - 1.
          Subsequent instances will be > 0 and just the duplicate number.  */
-      if ((pass->name && pass->name[0] != '*') || track_duplicates)
+      else if ((pass->name && pass->name[0] != '*') || track_duplicates)
         {
           pass->static_pass_number -= 1;
           new_pass->static_pass_number = -pass->static_pass_number;
@@ -953,12 +981,12 @@ make_pass_instance (struct opt_pass *pass, bool track_duplicates)
    in the list.  */
 
 static struct opt_pass **
-next_pass_1 (struct opt_pass **list, struct opt_pass *pass)
+next_pass_1 (struct opt_pass **list, struct opt_pass *pass, const char *suffix)
 {
   /* Every pass should have a name so that plugins can refer to them.  */
   gcc_assert (pass->name != NULL);
 
-  *list = make_pass_instance (pass, false);
+  *list = make_pass_instance (pass, false, suffix);
 
   return &(*list)->next;
 }
@@ -1008,7 +1036,7 @@ position_pass (struct register_pass_info *new_pass_info,
           struct opt_pass *new_pass;
           struct pass_list_node *new_pass_node;
 
-	  new_pass = make_pass_instance (new_pass_info->pass, true);
+	  new_pass = make_pass_instance (new_pass_info->pass, true, NULL);
 
           /* Insert the new pass instance based on the positioning op.  */
           switch (new_pass_info->pos_op)
@@ -1164,8 +1192,9 @@ void
 init_optimization_passes (void)
 {
   struct opt_pass **p;
+  const char *suffix = NULL;
 
-#define NEXT_PASS(PASS)  (p = next_pass_1 (p, &((PASS).pass)))
+#define NEXT_PASS(PASS)  (p = next_pass_1 (p, &((PASS).pass), suffix))
 
  /* All passes needed to lower the function into shape optimizers can
     operate on.  These passes are always run first on the function, but
@@ -1188,9 +1217,9 @@ init_optimization_passes (void)
   p = &all_small_ipa_passes;
   NEXT_PASS (pass_ipa_free_lang_data);
   NEXT_PASS (pass_ipa_function_and_variable_visibility);
-  NEXT_PASS (pass_early_local_passes);
+  NEXT_PASS (pass_early_local_passes_main);
     {
-      struct opt_pass **p = &pass_early_local_passes.pass.sub;
+      struct opt_pass **p = &pass_early_local_passes_main.pass.sub;
       NEXT_PASS (pass_fixup_cfg);
       NEXT_PASS (pass_init_datastructures);
       NEXT_PASS (pass_expand_omp);
@@ -1202,9 +1231,9 @@ init_optimization_passes (void)
       NEXT_PASS (pass_rebuild_cgraph_edges);
       NEXT_PASS (pass_inline_parameters);
       NEXT_PASS (pass_early_inline);
-      NEXT_PASS (pass_all_early_optimizations);
-	{
-	  struct opt_pass **p = &pass_all_early_optimizations.pass.sub;
+      NEXT_PASS (pass_early_optimizations_main);
+        {
+	  struct opt_pass **p = &pass_early_optimizations_main.pass.sub;
 	  NEXT_PASS (pass_remove_cgraph_callee_edges);
 	  NEXT_PASS (pass_rename_ssa_copies);
 	  NEXT_PASS (pass_ccp);
@@ -1222,18 +1251,61 @@ init_optimization_passes (void)
 	  NEXT_PASS (pass_early_ipa_sra);
 	  NEXT_PASS (pass_tail_recursion);
 	  NEXT_PASS (pass_convert_switch);
-          NEXT_PASS (pass_cleanup_eh);
-          NEXT_PASS (pass_profile);
-          NEXT_PASS (pass_local_pure_const);
+	  NEXT_PASS (pass_cleanup_eh);
+	  NEXT_PASS (pass_local_pure_const);
+	}
+      /* Define pass_early_optimizations_late, we run these optimizations
+	 strictly once per function.
+	 When not running iterative optimizations, schedule these
+	 optimizations together with pass_early_local_pass_main.
+	 Otherwise, run them after iterative optimizations.  */
+	{
+	  struct opt_pass **p = &pass_early_optimizations_late.pass.sub;
+	  NEXT_PASS (pass_profile);
 	  /* Split functions creates parts that are not run through
 	     early optimizations again.  It is thus good idea to do this
 	     late.  */
-          NEXT_PASS (pass_split_functions);
+	  NEXT_PASS (pass_split_functions);
+	  NEXT_PASS (pass_release_ssa_names);
 	}
-      NEXT_PASS (pass_release_ssa_names);
+      if (PARAM_VALUE (PARAM_EIPA_ITERATIONS) == 1)
+	NEXT_PASS (pass_early_optimizations_late);
       NEXT_PASS (pass_rebuild_cgraph_edges);
       NEXT_PASS (pass_inline_parameters);
     }
+
+  if (PARAM_VALUE (PARAM_EIPA_ITERATIONS) > 1)
+    {
+      /* Prepare and run several iterations of early_optimizations.  */
+      NEXT_PASS (pass_early_local_passes_iter);
+        {
+	  struct opt_pass **p = &pass_early_local_passes_iter.pass.sub;
+	  const char *suffix = "_iter";
+	  /* Fixup CFG as some of callees might have had their attributes
+	     (nothrow, pure, etc.) changed by pass_local_pure_const.
+	     Pass_fixup_cfg should always be followed by
+	     rebuild_cgraph_edges to remove outdated call-graph edges.  */
+	  NEXT_PASS (pass_fixup_cfg);
+	  NEXT_PASS (pass_rebuild_cgraph_edges);
+	  /* Run early inlining.  */
+	  NEXT_PASS (pass_inline_parameters);
+	  NEXT_PASS (pass_early_inline);
+	  /* Run the bulk of early optimizations.  */
+	  NEXT_PASS (pass_early_optimizations_main);
+	  /* Clean up and prepare to be inlined to some other function.  */
+	  NEXT_PASS (pass_rebuild_cgraph_edges);
+	  NEXT_PASS (pass_inline_parameters);
+	}
+      NEXT_PASS (pass_early_local_passes_late);
+        {
+	  struct opt_pass **p = &pass_early_local_passes_late.pass.sub;
+	  NEXT_PASS (pass_fixup_cfg);
+	  NEXT_PASS (pass_rebuild_cgraph_edges);
+	  NEXT_PASS (pass_early_optimizations_late);
+	  NEXT_PASS (pass_rebuild_cgraph_edges);
+	  NEXT_PASS (pass_inline_parameters);
+	}
+    }
   NEXT_PASS (pass_ipa_tree_profile);
     {
       struct opt_pass **p = &pass_ipa_tree_profile.pass.sub;
diff --git a/gcc/toplev.c b/gcc/toplev.c
index 86eed5d..4186157 100644
--- a/gcc/toplev.c
+++ b/gcc/toplev.c
@@ -1228,7 +1228,6 @@ general_init (const char *argv0)
   /* This must be done after global_init_params but before argument
      processing.  */
   init_ggc_heuristics();
-  init_optimization_passes ();
   statistics_early_init ();
   finish_params ();
 }
@@ -1989,6 +1988,8 @@ toplev_main (int argc, char **argv)
 		  save_decoded_options, save_decoded_options_count,
 		  UNKNOWN_LOCATION, global_dc);
 
+  init_optimization_passes ();
+
   handle_common_deferred_options ();
 
   init_local_tick ();
diff --git a/gcc/tree-optimize.c b/gcc/tree-optimize.c
index d7978b9..8516070 100644
--- a/gcc/tree-optimize.c
+++ b/gcc/tree-optimize.c
@@ -47,6 +47,7 @@ along with GCC; see the file COPYING3.  If not see
 #include "except.h"
 #include "plugin.h"
 #include "regset.h"	/* FIXME: For reg_obstack.  */
+#include "params.h"
 
 /* Gate: execute, or not, all of the non-trivial optimizations.  */
 
@@ -101,11 +102,11 @@ execute_all_early_local_passes (void)
   return 0;
 }
 
-struct simple_ipa_opt_pass pass_early_local_passes =
+struct simple_ipa_opt_pass pass_early_local_passes_main =
 {
  {
   SIMPLE_IPA_PASS,
-  "early_local_cleanups",		/* name */
+  "early_local_cleanups_main",		/* name */
   gate_all_early_local_passes,		/* gate */
   execute_all_early_local_passes,	/* execute */
   NULL,					/* sub */
@@ -120,6 +121,86 @@ struct simple_ipa_opt_pass pass_early_local_passes =
  }
 };
 
+static unsigned int
+execute_early_local_passes_iter (void)
+{
+  struct opt_pass *current = current_pass;
+  struct opt_pass *next = current_pass->next;
+  int i;
+
+  /* Don't recurse or wonder on to the next pass when running
+     execute_ipa_pass_list below.  */
+  current->execute = NULL;
+  current->next = NULL;
+
+  /* Run PARAM_EIPA_ITERATIONS-1 iterations of early optimizations.
+     We run these passes the first time as part of
+     pass_early_local_passes_main.  */
+  gcc_assert (PARAM_VALUE (PARAM_EIPA_ITERATIONS) > 1);
+  for (i = 1; i < PARAM_VALUE (PARAM_EIPA_ITERATIONS); ++i)
+    execute_ipa_pass_list (current);
+
+  /* Restore.  */
+  current->next = next;
+  current->execute = execute_early_local_passes_iter;
+  current_pass = current;
+
+  /* Tell execute_early_local_passes_for_current_function that it is OK to
+     run late optimizations now.  This also avoids running an extra iteration
+     of optimizations by execute_ipa_pass_list machinery.  */
+  current->sub = NULL;
+
+  return 0;
+}
+
+struct simple_ipa_opt_pass pass_early_local_passes_iter =
+{
+ {
+  SIMPLE_IPA_PASS,
+  "early_local_cleanups_iter",		/* name */
+  gate_all_early_local_passes,		/* gate */
+  execute_early_local_passes_iter,	/* execute */
+  NULL,					/* sub */
+  NULL,					/* next */
+  0,					/* static_pass_number */
+  TV_EARLY_LOCAL,			/* tv_id */
+  0,					/* properties_required */
+  0,					/* properties_provided */
+  0,					/* properties_destroyed */
+  0,					/* todo_flags_start */
+  TODO_remove_functions	 		/* todo_flags_finish */
+ }
+};
+
+struct simple_ipa_opt_pass pass_early_local_passes_late =
+{
+ {
+  SIMPLE_IPA_PASS,
+  "early_local_cleanups_late",		/* name */
+  gate_all_early_local_passes,		/* gate */
+  NULL,					/* execute */
+  NULL,					/* sub */
+  NULL,					/* next */
+  0,					/* static_pass_number */
+  TV_EARLY_LOCAL,			/* tv_id */
+  0,					/* properties_required */
+  0,					/* properties_provided */
+  0,					/* properties_destroyed */
+  0,					/* todo_flags_start */
+  TODO_remove_functions	 		/* todo_flags_finish */
+ }
+};
+
+void
+execute_early_local_passes_for_current_function (void)
+{
+  execute_pass_list (pass_early_local_passes_main.pass.sub);
+  /* Run early_local_passes_late if we are done with iterative optimizations
+     or not running them (iterative optimizations) at all.  */
+  if (pass_early_local_passes_iter.pass.sub == NULL)
+    execute_pass_list (pass_early_local_passes_late.pass.sub);
+}
+
 /* Gate: execute, or not, all of the non-trivial optimizations.  */
 
 static bool
@@ -130,11 +211,30 @@ gate_all_early_optimizations (void)
 	  && !seen_error ());
 }
 
-struct gimple_opt_pass pass_all_early_optimizations =
+struct gimple_opt_pass pass_early_optimizations_main =
+{
+ {
+  GIMPLE_PASS,
+  "early_optimizations_main",		/* name */
+  gate_all_early_optimizations,		/* gate */
+  NULL,					/* execute */
+  NULL,					/* sub */
+  NULL,					/* next */
+  0,					/* static_pass_number */
+  TV_NONE,				/* tv_id */
+  0,					/* properties_required */
+  0,					/* properties_provided */
+  0,					/* properties_destroyed */
+  0,					/* todo_flags_start */
+  0					/* todo_flags_finish */
+ }
+};
+
+struct gimple_opt_pass pass_early_optimizations_late =
 {
  {
   GIMPLE_PASS,
-  "early_optimizations",		/* name */
+  "early_optimizations_late",		/* name */
   gate_all_early_optimizations,		/* gate */
   NULL,					/* execute */
   NULL,					/* sub */
@@ -384,7 +484,7 @@ tree_lowering_passes (tree fn)
   bitmap_obstack_initialize (NULL);
   execute_pass_list (all_lowering_passes);
   if (optimize && cgraph_global_info_ready)
-    execute_pass_list (pass_early_local_passes.pass.sub);
+    execute_early_local_passes_for_current_function ();
   free_dominance_info (CDI_POST_DOMINATORS);
   free_dominance_info (CDI_DOMINATORS);
   compact_blocks ();
diff --git a/gcc/tree-pass.h b/gcc/tree-pass.h
index df1e24c..c1f44b9 100644
--- a/gcc/tree-pass.h
+++ b/gcc/tree-pass.h
@@ -455,7 +455,9 @@ extern struct simple_ipa_opt_pass pass_ipa_lower_emutls;
 extern struct simple_ipa_opt_pass pass_ipa_function_and_variable_visibility;
 extern struct simple_ipa_opt_pass pass_ipa_tree_profile;
 
-extern struct simple_ipa_opt_pass pass_early_local_passes;
+extern struct simple_ipa_opt_pass pass_early_local_passes_main;
+extern struct simple_ipa_opt_pass pass_early_local_passes_iter;
+extern struct simple_ipa_opt_pass pass_early_local_passes_late;
 
 extern struct ipa_opt_pass_d pass_ipa_whole_program_visibility;
 extern struct ipa_opt_pass_d pass_ipa_lto_gimple_out;
@@ -575,7 +577,8 @@ extern struct rtl_opt_pass pass_rtl_seqabstr;
 extern struct gimple_opt_pass pass_release_ssa_names;
 extern struct gimple_opt_pass pass_early_inline;
 extern struct gimple_opt_pass pass_inline_parameters;
-extern struct gimple_opt_pass pass_all_early_optimizations;
+extern struct gimple_opt_pass pass_early_optimizations_main;
+extern struct gimple_opt_pass pass_early_optimizations_late;
 extern struct gimple_opt_pass pass_update_address_taken;
 extern struct gimple_opt_pass pass_convert_switch;
 
@@ -638,6 +641,8 @@ extern void register_pass (struct register_pass_info *);
    directly in jump threading, and avoid peeling them next time.  */
 extern bool first_pass_instance;
 
+extern void execute_early_local_passes_for_current_function (void);
+
 /* Declare for plugins.  */
 extern void do_per_function_toporder (void (*) (void *), void *);
 
